"""Send message action for BlueBubbles."""

import logging
from typing import Any

from elizaos.types import Action, ActionExample, Content, Memory, State

logger = logging.getLogger(__name__)

BLUEBUBBLES_SERVICE_NAME = "bluebubbles"


async def validate(runtime: Any, message: Memory) -> bool:
    """Validates if the action can be executed."""
    service = runtime.get_service(BLUEBUBBLES_SERVICE_NAME)
    return service is not None and service.is_running


async def handler(
    runtime: Any,
    message: Memory,
    state: State | None = None,
    options: dict[str, Any] | None = None,
    callback: Any = None,
) -> Content | None:
    """Handles the send message action."""
    service = runtime.get_service(BLUEBUBBLES_SERVICE_NAME)

    if not service or not service.is_running:
        logger.error("BlueBubbles service is not available")
        if callback:
            callback(
                Content(
                    text="Sorry, the iMessage service is currently unavailable.",
                    error="BlueBubbles service not available",
                )
            )
        return None

    try:
        # Get the room to find the target
        room = await runtime.get_room(message.room_id)
        if not room or not room.channel_id:
            logger.error("No channel ID found for room")
            if callback:
                callback(
                    Content(
                        text="Unable to determine the message recipient.",
                        error="No channel ID",
                    )
                )
            return None

        text = message.content.text
        if not text or not text.strip():
            logger.warning("Empty message text, skipping send")
            return None

        reply_to = message.content.in_reply_to

        # Send the message
        guid = await service.send_message(room.channel_id, text, reply_to)

        logger.info("Sent BlueBubbles message: %s", guid)

        content = Content(
            text=text,
            source="bluebubbles",
            metadata={
                "messageGuid": guid,
                "chatGuid": room.channel_id,
            },
        )

        if callback:
            callback(content)

        return content

    except Exception as e:
        logger.error("Failed to send BlueBubbles message: %s", e)
        if callback:
            callback(
                Content(
                    text="Failed to send the iMessage. Please try again.",
                    error=str(e),
                )
            )
        return None


send_message_action = Action(
    name="SEND_BLUEBUBBLES_MESSAGE",
    description="Send a message via iMessage through BlueBubbles",
    similes=[
        "SEND_IMESSAGE",
        "TEXT_MESSAGE",
        "IMESSAGE_REPLY",
        "BLUEBUBBLES_SEND",
        "APPLE_MESSAGE",
    ],
    examples=[
        [
            ActionExample(
                name="{{user1}}",
                content=Content(
                    text="Can you send a message to John saying I'll be late?"
                ),
            ),
            ActionExample(
                name="{{agentName}}",
                content=Content(
                    text="I'll send that message to John for you.",
                    action="SEND_BLUEBUBBLES_MESSAGE",
                ),
            ),
        ],
        [
            ActionExample(
                name="{{user1}}",
                content=Content(text="Reply to this iMessage for me"),
            ),
            ActionExample(
                name="{{agentName}}",
                content=Content(
                    text="I'll compose and send a reply for you.",
                    action="SEND_BLUEBUBBLES_MESSAGE",
                ),
            ),
        ],
    ],
    validate=validate,
    handler=handler,
)



"""Tests for BlueBubbles configuration."""

import pytest

from elizaos_plugin_bluebubbles.config import (
    BlueBubblesConfig,
    is_handle_allowed,
    normalize_handle,
)
from elizaos_plugin_bluebubbles.types import DmPolicy


class TestNormalizeHandle:
    """Tests for normalize_handle function."""

    def test_normalize_phone_with_formatting(self):
        """Test normalizing a phone number with formatting."""
        assert normalize_handle("+1 (555) 123-4567") == "+15551234567"

    def test_normalize_phone_without_plus(self):
        """Test normalizing a phone number without plus prefix."""
        assert normalize_handle("555-123-4567") == "+5551234567"

    def test_normalize_international_phone(self):
        """Test normalizing an international phone number."""
        assert normalize_handle("+44 7700 900000") == "+447700900000"

    def test_normalize_email(self):
        """Test normalizing an email address."""
        assert normalize_handle("User@Example.COM") == "user@example.com"

    def test_normalize_email_with_whitespace(self):
        """Test normalizing an email with whitespace."""
        assert normalize_handle("  test@test.com  ") == "test@test.com"


class TestIsHandleAllowed:
    """Tests for is_handle_allowed function."""

    def test_open_policy_allows_all(self):
        """Test that open policy allows all handles."""
        assert is_handle_allowed("anyone", [], DmPolicy.OPEN)

    def test_disabled_policy_denies_all(self):
        """Test that disabled policy denies all handles."""
        assert not is_handle_allowed("anyone", [], DmPolicy.DISABLED)

    def test_pairing_empty_allows_first(self):
        """Test that pairing with empty allowlist allows first contact."""
        assert is_handle_allowed("first@contact.com", [], DmPolicy.PAIRING)

    def test_allowlist_matches_normalized(self):
        """Test that allowlist matches normalized handles."""
        allow_list = ["+15551234567"]
        assert is_handle_allowed("+1 (555) 123-4567", allow_list, DmPolicy.ALLOWLIST)

    def test_allowlist_denies_non_matching(self):
        """Test that allowlist denies non-matching handles."""
        allow_list = ["+15551234567"]
        assert not is_handle_allowed("+15559876543", allow_list, DmPolicy.ALLOWLIST)


class TestBlueBubblesConfig:
    """Tests for BlueBubblesConfig."""

    def test_valid_config(self):
        """Test creating a valid configuration."""
        config = BlueBubblesConfig(
            server_url="http://localhost:1234",
            password="password123",
        )
        assert config.server_url == "http://localhost:1234"
        assert config.password == "password123"

    def test_empty_url_fails(self):
        """Test that empty URL fails validation."""
        with pytest.raises(ValueError, match="Server URL is required"):
            BlueBubblesConfig(server_url="", password="password")

    def test_empty_password_fails(self):
        """Test that empty password fails validation."""
        with pytest.raises(ValueError, match="Password is required"):
            BlueBubblesConfig(server_url="http://localhost:1234", password="")

    def test_invalid_url_fails(self):
        """Test that invalid URL fails validation."""
        with pytest.raises(ValueError, match="must start with http"):
            BlueBubblesConfig(server_url="localhost:1234", password="password")

    def test_default_values(self):
        """Test default configuration values."""
        config = BlueBubblesConfig(
            server_url="http://localhost:1234",
            password="password",
        )
        assert config.webhook_path == "/webhooks/bluebubbles"
        assert config.dm_policy == DmPolicy.PAIRING
        assert config.send_read_receipts is True
        assert config.enabled is True



"""Type definitions for the BlueBubbles plugin."""

from enum import Enum
from typing import Any

from pydantic import BaseModel, Field


class DmPolicy(str, Enum):
    """DM policy options."""

    OPEN = "open"
    PAIRING = "pairing"
    ALLOWLIST = "allowlist"
    DISABLED = "disabled"


class GroupPolicy(str, Enum):
    """Group policy options."""

    OPEN = "open"
    ALLOWLIST = "allowlist"
    DISABLED = "disabled"


class BlueBubblesHandle(BaseModel):
    """BlueBubbles handle (contact)."""

    address: str
    service: str
    country: str | None = None
    original_rowid: int = Field(alias="originalROWID")
    uncanonicalized_id: str | None = Field(default=None, alias="uncanonicalizedId")

    class Config:
        populate_by_name = True


class BlueBubblesAttachment(BaseModel):
    """BlueBubbles attachment."""

    guid: str
    original_rowid: int = Field(alias="originalROWID")
    uti: str
    mime_type: str | None = Field(default=None, alias="mimeType")
    transfer_name: str = Field(alias="transferName")
    total_bytes: int = Field(alias="totalBytes")
    is_outgoing: bool = Field(alias="isOutgoing")
    hide_attachment: bool = Field(alias="hideAttachment")
    is_sticker: bool = Field(alias="isSticker")
    has_live_photo: bool = Field(alias="hasLivePhoto")
    height: int | None = None
    width: int | None = None
    metadata: dict[str, Any] | None = None

    class Config:
        populate_by_name = True


class BlueBubblesMessage(BaseModel):
    """BlueBubbles message."""

    guid: str
    text: str | None = None
    subject: str | None = None
    country: str | None = None
    handle: BlueBubblesHandle | None = None
    handle_id: int = Field(alias="handleId")
    other_handle: int = Field(alias="otherHandle")
    chats: list["BlueBubblesChat"] = []
    attachments: list[BlueBubblesAttachment] = []
    expressive_send_style_id: str | None = Field(default=None, alias="expressiveSendStyleId")
    date_created: int = Field(alias="dateCreated")
    date_read: int | None = Field(default=None, alias="dateRead")
    date_delivered: int | None = Field(default=None, alias="dateDelivered")
    is_from_me: bool = Field(alias="isFromMe")
    is_delayed: bool = Field(alias="isDelayed")
    is_auto_reply: bool = Field(alias="isAutoReply")
    is_system_message: bool = Field(alias="isSystemMessage")
    is_service_message: bool = Field(alias="isServiceMessage")
    is_forward: bool = Field(alias="isForward")
    is_archived: bool = Field(alias="isArchived")
    has_dd_results: bool = Field(alias="hasDdResults")
    has_payload_data: bool = Field(alias="hasPayloadData")
    thread_originator_guid: str | None = Field(default=None, alias="threadOriginatorGuid")
    thread_originator_part: str | None = Field(default=None, alias="threadOriginatorPart")
    associated_message_guid: str | None = Field(default=None, alias="associatedMessageGuid")
    associated_message_type: str | None = Field(default=None, alias="associatedMessageType")
    balloon_bundle_id: str | None = Field(default=None, alias="balloonBundleId")
    date_edited: int | None = Field(default=None, alias="dateEdited")
    error: int = 0
    item_type: int = Field(default=0, alias="itemType")
    group_title: str | None = Field(default=None, alias="groupTitle")
    group_action_type: int = Field(default=0, alias="groupActionType")
    payload_data: dict[str, Any] | None = Field(default=None, alias="payloadData")

    class Config:
        populate_by_name = True


class BlueBubblesChat(BaseModel):
    """BlueBubbles chat."""

    guid: str
    chat_identifier: str = Field(alias="chatIdentifier")
    display_name: str | None = Field(default=None, alias="displayName")
    participants: list[BlueBubblesHandle] = []
    last_message: BlueBubblesMessage | None = Field(default=None, alias="lastMessage")
    style: int = 0
    is_archived: bool = Field(default=False, alias="isArchived")
    is_filtered: bool = Field(default=False, alias="isFiltered")
    is_pinned: bool = Field(default=False, alias="isPinned")
    has_unread_messages: bool = Field(default=False, alias="hasUnreadMessages")

    class Config:
        populate_by_name = True


# Update forward reference
BlueBubblesMessage.model_rebuild()


class BlueBubblesServerInfo(BaseModel):
    """BlueBubbles server info."""

    os_version: str
    server_version: str
    private_api: bool
    proxy_service: str | None = None
    helper_connected: bool
    detected_icloud: str | None = None


class SendMessageOptions(BaseModel):
    """Options for sending a message."""

    temp_guid: str | None = None
    method: str | None = "apple-script"
    subject: str | None = None
    effect_id: str | None = None
    part_index: int | None = None
    dd_scan: bool | None = None


class SendMessageResult(BaseModel):
    """Result of sending a message."""

    guid: str
    temp_guid: str | None = None
    status: str = "sent"
    date_created: int
    text: str
    error: str | None = None


class BlueBubblesProbeResult(BaseModel):
    """Result of probing the BlueBubbles server."""

    ok: bool
    server_version: str | None = None
    os_version: str | None = None
    private_api_enabled: bool | None = None
    helper_connected: bool | None = None
    error: str | None = None


class BlueBubblesChatState(BaseModel):
    """State information for a BlueBubbles chat."""

    chat_guid: str
    chat_identifier: str
    is_group: bool
    participants: list[str]
    display_name: str | None = None
    last_message_at: int | None = None
    has_unread: bool = False


class BlueBubblesWebhookPayload(BaseModel):
    """Webhook payload from BlueBubbles."""

    type: str
    data: dict[str, Any]



"""BlueBubbles service implementation."""

import logging
from typing import Any

from elizaos_plugin_bluebubbles.client import BlueBubblesClient
from elizaos_plugin_bluebubbles.config import (
    BlueBubblesConfig,
    get_config_from_env,
    is_group_handle_allowed,
    is_handle_allowed,
)
from elizaos_plugin_bluebubbles.types import (
    BlueBubblesChat,
    BlueBubblesChatState,
    BlueBubblesMessage,
    BlueBubblesWebhookPayload,
)

logger = logging.getLogger(__name__)

BLUEBUBBLES_SERVICE_NAME = "bluebubbles"


class BlueBubblesService:
    """BlueBubbles service for ElizaOS."""

    service_type = BLUEBUBBLES_SERVICE_NAME
    capability_description = (
        "The agent is able to send and receive iMessages via BlueBubbles"
    )

    def __init__(self) -> None:
        """Initializes the service."""
        self.client: BlueBubblesClient | None = None
        self.config: BlueBubblesConfig | None = None
        self.known_chats: dict[str, BlueBubblesChat] = {}
        self._is_running = False

    @property
    def is_running(self) -> bool:
        """Returns whether the service is running."""
        return self._is_running

    @property
    def webhook_path(self) -> str:
        """Returns the webhook path."""
        if self.config:
            return self.config.webhook_path
        return "/webhooks/bluebubbles"

    async def start(self, runtime: Any = None) -> "BlueBubblesService":
        """Starts the service."""
        self.config = get_config_from_env()

        if not self.config:
            logger.warning(
                "BlueBubbles configuration not available - service unavailable"
            )
            return self

        if not self.config.enabled:
            logger.info("BlueBubbles plugin is disabled via configuration")
            return self

        self.client = BlueBubblesClient(self.config)

        # Probe the server
        probe_result = await self.client.probe()

        if not probe_result.ok:
            logger.error(
                "Failed to connect to BlueBubbles server: %s", probe_result.error
            )
            return self

        logger.info(
            "Connected to BlueBubbles server v%s on macOS %s",
            probe_result.server_version,
            probe_result.os_version,
        )

        if probe_result.private_api_enabled:
            logger.info(
                "BlueBubbles Private API enabled - edit and unsend available"
            )

        # Load initial chats
        try:
            chats = await self.client.list_chats(100)
            for chat in chats:
                self.known_chats[chat.guid] = chat
            logger.info("Loaded %d BlueBubbles chats", len(self.known_chats))
        except Exception as e:
            logger.warning("Failed to load initial chats: %s", e)

        self._is_running = True
        logger.info("BlueBubbles service started")

        return self

    async def stop(self, runtime: Any = None) -> None:
        """Stops the service."""
        self._is_running = False
        if self.client:
            await self.client.close()
        logger.info("BlueBubbles service stopped")

    async def handle_webhook(self, payload: BlueBubblesWebhookPayload) -> None:
        """Handles an incoming webhook payload."""
        if not self.config or not self.client:
            logger.warning("Webhook received but service not configured")
            return

        event_type = payload.type

        if event_type == "new-message":
            message = BlueBubblesMessage(**payload.data)
            await self._handle_incoming_message(message)
        elif event_type == "updated-message":
            message = BlueBubblesMessage(**payload.data)
            await self._handle_message_update(message)
        elif event_type == "chat-updated":
            chat = BlueBubblesChat(**payload.data)
            await self._handle_chat_update(chat)
        elif event_type in ("typing-indicator", "read-receipt"):
            logger.debug("BlueBubbles %s: %s", event_type, payload.data)
        else:
            logger.debug("Unhandled BlueBubbles event: %s", event_type)

    async def _handle_incoming_message(self, message: BlueBubblesMessage) -> None:
        """Handles an incoming message."""
        if not self.config:
            return

        # Skip outgoing messages
        if message.is_from_me:
            return

        # Skip system messages
        if message.is_system_message:
            return

        if not message.chats:
            logger.warning("Message without chat info: %s", message.guid)
            return

        chat = message.chats[0]
        is_group = len(chat.participants) > 1
        sender_handle = message.handle.address if message.handle else ""

        # Check access policies
        if is_group:
            if not is_group_handle_allowed(
                sender_handle,
                self.config.group_allow_from,
                self.config.group_policy,
            ):
                logger.debug("Ignoring from %s - not in group allowlist", sender_handle)
                return
        else:
            if not is_handle_allowed(
                sender_handle, self.config.allow_from, self.config.dm_policy
            ):
                logger.debug("Ignoring from %s - not in DM allowlist", sender_handle)
                return

        # Mark as read if configured
        if self.config.send_read_receipts and self.client:
            try:
                await self.client.mark_chat_read(chat.guid)
            except Exception as e:
                logger.debug("Failed to mark chat as read: %s", e)

        logger.info(
            "Received message from %s in chat %s: %s",
            sender_handle,
            chat.guid,
            message.text or "[no text]",
        )

    async def _handle_message_update(self, message: BlueBubblesMessage) -> None:
        """Handles a message update."""
        if message.date_edited:
            logger.debug("Message %s was edited", message.guid)

    async def _handle_chat_update(self, chat: BlueBubblesChat) -> None:
        """Handles a chat update."""
        logger.debug(
            "Chat %s updated: %s",
            chat.guid,
            chat.display_name or chat.chat_identifier,
        )
        self.known_chats[chat.guid] = chat

    async def send_message(
        self, target: str, text: str, reply_to_id: str | None = None
    ) -> str:
        """Sends a message to a target."""
        if not self.client:
            raise RuntimeError("BlueBubbles client not initialized")

        chat_guid = await self.client.resolve_target(target)
        result = await self.client.send_message(chat_guid, text)

        return result.guid

    async def get_chat_state(self, chat_guid: str) -> BlueBubblesChatState | None:
        """Gets the state for a chat."""
        chat = self.known_chats.get(chat_guid)

        if not chat and self.client:
            try:
                chat = await self.client.get_chat(chat_guid)
                self.known_chats[chat.guid] = chat
            except Exception:
                return None

        if not chat:
            return None

        return self._chat_to_state(chat)

    def _chat_to_state(self, chat: BlueBubblesChat) -> BlueBubblesChatState:
        """Converts a chat to a chat state."""
        return BlueBubblesChatState(
            chat_guid=chat.guid,
            chat_identifier=chat.chat_identifier,
            is_group=len(chat.participants) > 1,
            participants=[p.address for p in chat.participants],
            display_name=chat.display_name,
            last_message_at=(
                chat.last_message.date_created if chat.last_message else None
            ),
            has_unread=chat.has_unread_messages,
        )



"""
Chat context provider for the BlueBubbles plugin.
"""

from typing import Protocol

from ..service import BlueBubblesService
from ..types import BLUEBUBBLES_SERVICE_NAME, extract_handle_from_chat_guid


class IAgentRuntime(Protocol):
    """Agent runtime protocol."""

    def get_service(self, service_type: str) -> BlueBubblesService | None:
        """Get a service by type."""
        ...


async def chat_context_get(
    runtime: IAgentRuntime,
    message: dict,
    state: dict | None = None,
) -> dict:
    """Get chat context."""
    # Only provide context for BlueBubbles messages
    content = message.get("content", {})
    if content.get("source") != "bluebubbles":
        return {
            "data": {},
            "values": {},
            "text": "",
        }

    bb_service = runtime.get_service(BLUEBUBBLES_SERVICE_NAME)

    if not bb_service or not bb_service.is_connected():
        return {
            "data": {"connected": False},
            "values": {"connected": False},
            "text": "",
        }

    state = state or {}
    agent_name = state.get("agentName", "The agent")
    state_data = state.get("data", {})

    chat_guid = state_data.get("chatGuid")
    handle = state_data.get("handle")
    display_name = state_data.get("displayName")

    # Determine chat type from GUID
    chat_type = "direct"
    chat_description = ""

    if chat_guid:
        if ";+;" in chat_guid:
            chat_type = "group"
            chat_description = f'group chat "{display_name}"' if display_name else "a group chat"
        else:
            extracted_handle = extract_handle_from_chat_guid(chat_guid)
            if extracted_handle:
                chat_description = f"direct message with {extracted_handle}"
            elif handle:
                chat_description = f"direct message with {handle}"
            else:
                chat_description = "a direct message"
    elif handle:
        chat_description = f"direct message with {handle}"
    else:
        chat_description = "an iMessage conversation"

    response_text = (
        f"{agent_name} is chatting via iMessage (BlueBubbles) in {chat_description}. "
        "This channel supports reactions, effects (slam, balloons, confetti, etc.), "
        "editing, and replying to messages."
    )

    return {
        "data": {
            "chatGuid": chat_guid,
            "handle": handle,
            "displayName": display_name,
            "chatType": chat_type,
            "connected": True,
            "platform": "bluebubbles",
            "supportsReactions": True,
            "supportsEffects": True,
            "supportsEdit": True,
            "supportsReply": True,
        },
        "values": {
            "chatGuid": chat_guid,
            "handle": handle,
            "displayName": display_name,
            "chatType": chat_type,
        },
        "text": response_text,
    }


chat_context_provider = {
    "name": "bluebubblesChatContext",
    "description": "Provides information about the current BlueBubbles/iMessage chat context",
    "get": chat_context_get,
}



"""
Send reaction action for the BlueBubbles plugin.
"""

import logging
from typing import Protocol

from ..service import BlueBubblesService
from ..types import BLUEBUBBLES_SERVICE_NAME

logger = logging.getLogger(__name__)


class IAgentRuntime(Protocol):
    """Agent runtime protocol."""

    def get_service(self, service_type: str) -> BlueBubblesService | None:
        """Get a service by type."""
        ...


SEND_REACTION_TEMPLATE = """# Task: Extract BlueBubbles reaction parameters

Based on the conversation, determine what reaction to add or remove.

Recent conversation:
{{recentMessages}}

Extract the following:
1. emoji: The emoji reaction to add (heart, thumbsup, thumbsdown, haha, \
exclamation, question, or any emoji)
2. messageId: The message ID to react to (or "last" for the last message)
3. remove: true to remove the reaction, false to add it

Respond with a JSON object:
```json
{
  "emoji": "❤️",
  "messageId": "last",
  "remove": false
}
```
"""


async def send_reaction_handler(
    runtime: IAgentRuntime,
    message: dict,
    state: dict | None = None,
    options: dict | None = None,
    callback=None,
) -> dict:
    """Handle send reaction action."""
    bb_service = runtime.get_service(BLUEBUBBLES_SERVICE_NAME)

    if not bb_service or not bb_service.is_connected():
        if callback:
            callback({"text": "BlueBubbles service is not available.", "source": "bluebubbles"})
        return {"success": False, "error": "BlueBubbles service not available"}

    # Get chat context
    state_data = (state or {}).get("data", {})
    chat_guid = state_data.get("chatGuid")
    message_guid = state_data.get("lastMessageGuid")

    if not chat_guid:
        if callback:
            callback({
                "text": "I couldn't determine the chat to react in.",
                "source": "bluebubbles"
            })
        return {"success": False, "error": "Could not determine chat"}

    if not message_guid:
        if callback:
            callback({"text": "I couldn't find the message to react to.", "source": "bluebubbles"})
        return {"success": False, "error": "Could not find message to react to"}

    # Default emoji (would be extracted by LLM in real usage)
    emoji = "❤️"
    remove = False

    # Send reaction
    result = await bb_service.send_reaction(chat_guid, message_guid, emoji, remove)

    if not result.success:
        action_text = "remove" if remove else "add"
        if callback:
            callback({
                "text": f"Failed to {action_text} reaction: {result.error}",
                "source": "bluebubbles",
            })
        return {"success": False, "error": result.error}

    logger.debug(f"{'Removed' if remove else 'Added'} reaction {emoji} on {message_guid}")

    content = message.get("content", {})
    if callback:
        callback({
            "text": "Reaction removed." if remove else f"Reacted with {emoji}.",
            "source": content.get("source", "bluebubbles"),
        })

    return {
        "success": True,
        "data": {
            "emoji": emoji,
            "messageGuid": message_guid,
            "removed": remove,
        },
    }


def _validate_reaction(runtime, message, state=None):
    return message.get("content", {}).get("source") == "bluebubbles"


send_reaction_action = {
    "name": "BLUEBUBBLES_SEND_REACTION",
    "similes": ["BLUEBUBBLES_REACT", "BB_REACTION", "IMESSAGE_REACT"],
    "description": "Add or remove a reaction on a message via BlueBubbles",
    "validate": _validate_reaction,
    "handler": send_reaction_handler,
    "examples": [
        [
            {"name": "{{user1}}", "content": {"text": "React to that message with a heart"}},
            {
                "name": "{{agent}}",
                "content": {
                    "text": "I'll add a heart reaction.",
                    "actions": ["BLUEBUBBLES_SEND_REACTION"],
                },
            },
        ],
    ],
}



"""Chat state provider for BlueBubbles."""

import logging
from datetime import datetime
from typing import Any

from elizaos.types import Memory, Provider, State

from elizaos_plugin_bluebubbles.types import BlueBubblesChatState

logger = logging.getLogger(__name__)

BLUEBUBBLES_SERVICE_NAME = "bluebubbles"


async def get_chat_state(
    runtime: Any,
    message: Memory,
    state: State | None = None,
) -> str:
    """Gets the chat state for BlueBubbles."""
    service = runtime.get_service(BLUEBUBBLES_SERVICE_NAME)

    if not service or not service.is_running:
        return ""

    try:
        room = await runtime.get_room(message.room_id)
        if not room or not room.channel_id:
            return ""

        # Only provide state for BlueBubbles channels
        if room.source != "bluebubbles":
            return ""

        chat_state = await service.get_chat_state(room.channel_id)
        if not chat_state:
            return ""

        return _format_chat_state(chat_state)

    except Exception as e:
        logger.debug("Failed to get BlueBubbles chat state: %s", e)
        return ""


def _format_chat_state(state: BlueBubblesChatState) -> str:
    """Formats the chat state for inclusion in prompts."""
    lines = [
        "# iMessage Chat Context (BlueBubbles)",
        "",
        f"- Chat Type: {'Group Chat' if state.is_group else 'Direct Message'}",
    ]

    if state.display_name:
        lines.append(f"- Chat Name: {state.display_name}")

    if state.is_group:
        lines.append(f"- Participants: {', '.join(state.participants)}")
    else:
        contact = state.participants[0] if state.participants else state.chat_identifier
        lines.append(f"- Contact: {contact}")

    if state.last_message_at:
        dt = datetime.fromtimestamp(state.last_message_at / 1000)
        lines.append(f"- Last Message: {dt.strftime('%Y-%m-%d %H:%M:%S')}")

    if state.has_unread:
        lines.append("- Has Unread Messages: Yes")

    lines.append("")
    lines.append(
        "Note: This conversation is happening through iMessage. "
        "Be conversational and friendly."
    )

    return "\n".join(lines)


chat_state_provider = Provider(
    name="BLUEBUBBLES_CHAT_STATE",
    description="Provides information about the current BlueBubbles/iMessage chat context",
    get=get_chat_state,
)



"""Tests for BlueBubbles action handlers."""

from unittest.mock import AsyncMock, MagicMock

import pytest

from elizaos_plugin_bluebubbles.actions.send_message import (
    handler as send_message_handler,
)
from elizaos_plugin_bluebubbles.actions.send_message import (
    send_message_action,
)
from elizaos_plugin_bluebubbles.actions.send_message import (
    validate as send_message_validate,
)
from elizaos_plugin_bluebubbles.actions.send_reaction import (
    _validate_reaction,
    send_reaction_action,
    send_reaction_handler,
)

# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------

def _make_runtime(*, service=None):
    """Create a mock runtime with an optional BlueBubbles service."""
    rt = MagicMock()
    rt.get_service = MagicMock(return_value=service)
    return rt


def _make_bb_service(*, is_running=True, is_connected=True, send_guid="msg-guid-123"):
    """Create a mock BlueBubbles service."""
    svc = AsyncMock()
    svc.is_running = is_running
    svc.is_connected = MagicMock(return_value=is_connected)

    svc.send_message = AsyncMock(return_value=send_guid)
    svc.send_reaction = AsyncMock(
        return_value=MagicMock(success=True, error=None)
    )

    return svc


def _make_room(channel_id="chat_guid:iMessage;-;+15551234567"):
    """Create a mock room."""
    room = MagicMock()
    room.channel_id = channel_id
    return room


def _make_memory(text="Hello world", room_id="room-1", in_reply_to=None, source="bluebubbles"):
    """Create a mock Memory object."""
    mem = MagicMock()
    mem.room_id = room_id
    mem.content = MagicMock()
    mem.content.text = text
    mem.content.in_reply_to = in_reply_to
    mem.content.source = source
    return mem


# ---------------------------------------------------------------------------
# send_message – validate
# ---------------------------------------------------------------------------

class TestSendMessageValidate:
    """Tests for send_message_action.validate."""

    @pytest.mark.asyncio
    async def test_valid_when_service_present_and_running(self):
        svc = _make_bb_service()
        rt = _make_runtime(service=svc)
        msg = _make_memory()
        assert await send_message_validate(rt, msg) is True

    @pytest.mark.asyncio
    async def test_invalid_when_service_missing(self):
        rt = _make_runtime(service=None)
        msg = _make_memory()
        assert await send_message_validate(rt, msg) is False

    @pytest.mark.asyncio
    async def test_invalid_when_service_not_running(self):
        svc = _make_bb_service(is_running=False)
        rt = _make_runtime(service=svc)
        msg = _make_memory()
        assert await send_message_validate(rt, msg) is False


# ---------------------------------------------------------------------------
# send_message – handler
# ---------------------------------------------------------------------------

class TestSendMessageHandler:
    """Tests for send_message_action.handler."""

    @pytest.mark.asyncio
    async def test_successful_send(self):
        svc = _make_bb_service()
        rt = _make_runtime(service=svc)
        rt.get_room = AsyncMock(return_value=_make_room())
        msg = _make_memory(text="Hello from agent!")

        result = await send_message_handler(rt, msg)

        assert result is not None
        assert result.text == "Hello from agent!"
        assert result.source == "bluebubbles"
        svc.send_message.assert_awaited_once()

    @pytest.mark.asyncio
    async def test_returns_none_when_service_unavailable(self):
        svc = _make_bb_service(is_running=False)
        rt = _make_runtime(service=svc)
        callback = MagicMock()

        result = await send_message_handler(rt, _make_memory(), callback=callback)

        assert result is None
        callback.assert_called_once()

    @pytest.mark.asyncio
    async def test_returns_none_when_no_room(self):
        svc = _make_bb_service()
        rt = _make_runtime(service=svc)
        rt.get_room = AsyncMock(return_value=None)
        callback = MagicMock()

        result = await send_message_handler(rt, _make_memory(), callback=callback)

        assert result is None
        callback.assert_called_once()

    @pytest.mark.asyncio
    async def test_returns_none_for_empty_text(self):
        svc = _make_bb_service()
        rt = _make_runtime(service=svc)
        rt.get_room = AsyncMock(return_value=_make_room())

        result = await send_message_handler(rt, _make_memory(text=""))

        assert result is None

    @pytest.mark.asyncio
    async def test_returns_none_for_whitespace_text(self):
        svc = _make_bb_service()
        rt = _make_runtime(service=svc)
        rt.get_room = AsyncMock(return_value=_make_room())

        result = await send_message_handler(rt, _make_memory(text="   "))

        assert result is None


# ---------------------------------------------------------------------------
# send_reaction – validate
# ---------------------------------------------------------------------------

class TestSendReactionValidate:
    """Tests for send_reaction validate."""

    def test_accepts_bluebubbles_source(self):
        msg = {"content": {"source": "bluebubbles"}}
        assert _validate_reaction(None, msg) is True

    def test_rejects_non_bluebubbles_source(self):
        msg = {"content": {"source": "discord"}}
        assert _validate_reaction(None, msg) is False

    def test_rejects_missing_source(self):
        msg = {"content": {}}
        assert _validate_reaction(None, msg) is False


# ---------------------------------------------------------------------------
# send_reaction – handler
# ---------------------------------------------------------------------------

class TestSendReactionHandler:
    """Tests for send_reaction_handler."""

    @pytest.mark.asyncio
    async def test_successful_reaction(self):
        svc = _make_bb_service()
        rt = _make_runtime(service=svc)
        msg = {"content": {"source": "bluebubbles"}}
        state = {"data": {"chatGuid": "iMessage;-;+15551234567", "lastMessageGuid": "msg-1"}}

        result = await send_reaction_handler(rt, msg, state=state)

        assert result["success"] is True
        assert result["data"]["emoji"] == "❤️"
        svc.send_reaction.assert_awaited_once()

    @pytest.mark.asyncio
    async def test_fails_when_service_not_connected(self):
        svc = _make_bb_service(is_connected=False)
        rt = _make_runtime(service=svc)

        result = await send_reaction_handler(rt, {}, state=None)

        assert result["success"] is False
        assert "not available" in result["error"]

    @pytest.mark.asyncio
    async def test_fails_when_no_chat_guid(self):
        svc = _make_bb_service()
        rt = _make_runtime(service=svc)
        state = {"data": {"lastMessageGuid": "msg-1"}}

        result = await send_reaction_handler(rt, {}, state=state)

        assert result["success"] is False
        assert "chat" in result["error"].lower()

    @pytest.mark.asyncio
    async def test_fails_when_no_message_guid(self):
        svc = _make_bb_service()
        rt = _make_runtime(service=svc)
        state = {"data": {"chatGuid": "iMessage;-;+15551234567"}}

        result = await send_reaction_handler(rt, {}, state=state)

        assert result["success"] is False
        assert "message" in result["error"].lower()


# ---------------------------------------------------------------------------
# Action definition structure
# ---------------------------------------------------------------------------

class TestActionDefinitions:
    """Tests for action dict/object shapes."""

    def test_send_message_action_name(self):
        assert send_message_action.name == "SEND_BLUEBUBBLES_MESSAGE"

    def test_send_message_has_similes(self):
        assert len(send_message_action.similes) >= 3

    def test_send_reaction_action_name(self):
        assert send_reaction_action["name"] == "BLUEBUBBLES_SEND_REACTION"

    def test_send_reaction_has_similes(self):
        assert "BLUEBUBBLES_REACT" in send_reaction_action["similes"]

    def test_send_reaction_has_examples(self):
        assert len(send_reaction_action["examples"]) >= 1



"""Configuration for the BlueBubbles plugin."""

import os
import re
from typing import Self

from pydantic import BaseModel, Field, model_validator

from elizaos_plugin_bluebubbles.types import DmPolicy, GroupPolicy

DEFAULT_WEBHOOK_PATH = "/webhooks/bluebubbles"


class BlueBubblesConfig(BaseModel):
    """BlueBubbles plugin configuration."""

    server_url: str = Field(description="BlueBubbles server URL")
    password: str = Field(description="Server password")
    webhook_path: str = Field(
        default=DEFAULT_WEBHOOK_PATH, description="Webhook path for receiving messages"
    )
    dm_policy: DmPolicy = Field(default=DmPolicy.PAIRING, description="DM policy")
    group_policy: GroupPolicy = Field(default=GroupPolicy.ALLOWLIST, description="Group policy")
    allow_from: list[str] = Field(default_factory=list, description="Allow list for DMs")
    group_allow_from: list[str] = Field(
        default_factory=list, description="Allow list for groups"
    )
    send_read_receipts: bool = Field(default=True, description="Send read receipts")
    enabled: bool = Field(default=True, description="Whether the plugin is enabled")

    @model_validator(mode="after")
    def validate_config(self) -> Self:
        """Validates the configuration."""
        if not self.server_url:
            raise ValueError("Server URL is required")
        if not self.password:
            raise ValueError("Password is required")

        # Validate URL format
        if not self.server_url.startswith(("http://", "https://")):
            raise ValueError("Server URL must start with http:// or https://")

        return self


def get_config_from_env() -> BlueBubblesConfig | None:
    """Gets BlueBubbles configuration from environment variables."""
    server_url = os.getenv("BLUEBUBBLES_SERVER_URL")
    password = os.getenv("BLUEBUBBLES_PASSWORD")

    if not server_url or not password:
        return None

    def parse_allow_list(raw: str | None) -> list[str]:
        if not raw:
            return []
        return [s.strip() for s in raw.split(",") if s.strip()]

    dm_policy_str = os.getenv("BLUEBUBBLES_DM_POLICY", "pairing").lower()
    group_policy_str = os.getenv("BLUEBUBBLES_GROUP_POLICY", "allowlist").lower()

    dm_policy_map = {
        "open": DmPolicy.OPEN,
        "pairing": DmPolicy.PAIRING,
        "allowlist": DmPolicy.ALLOWLIST,
        "disabled": DmPolicy.DISABLED,
    }

    group_policy_map = {
        "open": GroupPolicy.OPEN,
        "allowlist": GroupPolicy.ALLOWLIST,
        "disabled": GroupPolicy.DISABLED,
    }

    return BlueBubblesConfig(
        server_url=server_url,
        password=password,
        webhook_path=os.getenv("BLUEBUBBLES_WEBHOOK_PATH", DEFAULT_WEBHOOK_PATH),
        dm_policy=dm_policy_map.get(dm_policy_str, DmPolicy.PAIRING),
        group_policy=group_policy_map.get(group_policy_str, GroupPolicy.ALLOWLIST),
        allow_from=parse_allow_list(os.getenv("BLUEBUBBLES_ALLOW_FROM")),
        group_allow_from=parse_allow_list(os.getenv("BLUEBUBBLES_GROUP_ALLOW_FROM")),
        send_read_receipts=os.getenv("BLUEBUBBLES_SEND_READ_RECEIPTS", "true").lower() != "false",
        enabled=os.getenv("BLUEBUBBLES_ENABLED", "true").lower() != "false",
    )


def normalize_handle(handle: str) -> str:
    """Normalizes a phone number or email handle."""
    trimmed = handle.strip()

    # If it looks like an email, lowercase it
    if "@" in trimmed and not trimmed.startswith("+"):
        return trimmed.lower()

    # For phone numbers, strip non-digits except leading +
    starts_with_plus = trimmed.startswith("+")
    digits = re.sub(r"\D", "", trimmed)

    # Add + prefix if it was there or if we have 10+ digits (assume international)
    if starts_with_plus or len(digits) >= 10:
        return f"+{digits}"

    return digits


def is_handle_allowed(handle: str, allow_list: list[str], policy: DmPolicy) -> bool:
    """Checks if a handle is allowed based on policy."""
    if policy == DmPolicy.OPEN:
        return True

    if policy == DmPolicy.DISABLED:
        return False

    if policy in (DmPolicy.PAIRING, DmPolicy.ALLOWLIST):
        if not allow_list and policy == DmPolicy.PAIRING:
            # Pairing mode with empty allow list allows first contact
            return True

        normalized = normalize_handle(handle)
        return any(normalize_handle(allowed) == normalized for allowed in allow_list)

    return False


def is_group_handle_allowed(handle: str, allow_list: list[str], policy: GroupPolicy) -> bool:
    """Checks if a handle is allowed for groups."""
    if policy == GroupPolicy.OPEN:
        return True

    if policy == GroupPolicy.DISABLED:
        return False

    if policy == GroupPolicy.ALLOWLIST:
        normalized = normalize_handle(handle)
        return any(normalize_handle(allowed) == normalized for allowed in allow_list)

    return False



"""
BlueBubbles iMessage bridge plugin for ElizaOS.

This plugin provides iMessage integration via the BlueBubbles macOS app and REST API.
"""

from elizaos_plugin_bluebubbles.actions import send_message_action
from elizaos_plugin_bluebubbles.client import BlueBubblesClient
from elizaos_plugin_bluebubbles.config import BlueBubblesConfig, get_config_from_env
from elizaos_plugin_bluebubbles.providers import chat_state_provider
from elizaos_plugin_bluebubbles.service import BlueBubblesService
from elizaos_plugin_bluebubbles.types import (
    BlueBubblesAttachment,
    BlueBubblesChat,
    BlueBubblesChatState,
    BlueBubblesHandle,
    BlueBubblesMessage,
    BlueBubblesProbeResult,
    BlueBubblesServerInfo,
    DmPolicy,
    GroupPolicy,
    SendMessageOptions,
    SendMessageResult,
)

__all__ = [
    # Client
    "BlueBubblesClient",
    # Config
    "BlueBubblesConfig",
    "get_config_from_env",
    # Service
    "BlueBubblesService",
    # Types
    "BlueBubblesAttachment",
    "BlueBubblesChat",
    "BlueBubblesChatState",
    "BlueBubblesHandle",
    "BlueBubblesMessage",
    "BlueBubblesProbeResult",
    "BlueBubblesServerInfo",
    "DmPolicy",
    "GroupPolicy",
    "SendMessageOptions",
    "SendMessageResult",
    # Actions
    "send_message_action",
    # Providers
    "chat_state_provider",
]

BLUEBUBBLES_SERVICE_NAME = "bluebubbles"


def create_plugin():
    """Creates the BlueBubbles plugin with all components."""
    from elizaos.types import Plugin

    return Plugin(
        name=BLUEBUBBLES_SERVICE_NAME,
        description="BlueBubbles iMessage bridge plugin for ElizaOS agents",
        services=[BlueBubblesService],
        actions=[send_message_action],
        providers=[chat_state_provider],
    )



"""BlueBubbles API client."""

import asyncio
import logging
from typing import Any
from urllib.parse import quote

import httpx

from elizaos_plugin_bluebubbles.config import BlueBubblesConfig
from elizaos_plugin_bluebubbles.types import (
    BlueBubblesChat,
    BlueBubblesMessage,
    BlueBubblesProbeResult,
    BlueBubblesServerInfo,
    SendMessageOptions,
    SendMessageResult,
)

logger = logging.getLogger(__name__)


class BlueBubblesClient:
    """Client for interacting with the BlueBubbles server."""

    def __init__(self, config: BlueBubblesConfig):
        """Initializes the client."""
        self.base_url = config.server_url.rstrip("/")
        self.password = config.password
        self._client = httpx.AsyncClient(timeout=30.0)

    async def close(self) -> None:
        """Closes the HTTP client."""
        await self._client.aclose()

    def _build_url(self, endpoint: str) -> str:
        """Builds a URL with password authentication."""
        separator = "&" if "?" in endpoint else "?"
        return f"{self.base_url}{endpoint}{separator}password={quote(self.password)}"

    async def _get(self, endpoint: str) -> dict[str, Any]:
        """Makes a GET request."""
        url = self._build_url(endpoint)
        logger.debug("GET %s", endpoint)

        response = await self._client.get(url)

        if not response.is_success:
            raise Exception(
                f"BlueBubbles API error ({response.status_code}): {response.text}"
            )

        return response.json()["data"]

    async def _post(self, endpoint: str, body: dict[str, Any]) -> dict[str, Any]:
        """Makes a POST request."""
        url = self._build_url(endpoint)
        logger.debug("POST %s", endpoint)

        response = await self._client.post(url, json=body)

        if not response.is_success:
            raise Exception(
                f"BlueBubbles API error ({response.status_code}): {response.text}"
            )

        return response.json()["data"]

    async def probe(self, timeout_ms: int = 5000) -> BlueBubblesProbeResult:
        """Probes the server to check connectivity."""
        try:
            async with asyncio.timeout(timeout_ms / 1000):
                url = self._build_url("/api/v1/server/info")
                response = await self._client.get(url)

                if not response.is_success:
                    return BlueBubblesProbeResult(
                        ok=False, error=f"HTTP {response.status_code}"
                    )

                data = response.json()["data"]
                info = BlueBubblesServerInfo(**data)

                return BlueBubblesProbeResult(
                    ok=True,
                    server_version=info.server_version,
                    os_version=info.os_version,
                    private_api_enabled=info.private_api,
                    helper_connected=info.helper_connected,
                )
        except asyncio.TimeoutError:
            return BlueBubblesProbeResult(ok=False, error="Connection timeout")
        except Exception as e:
            return BlueBubblesProbeResult(ok=False, error=str(e))

    async def send_message(
        self,
        chat_guid: str,
        text: str,
        options: SendMessageOptions | None = None,
    ) -> SendMessageResult:
        """Sends a text message."""
        opts = options or SendMessageOptions()

        body: dict[str, Any] = {
            "chatGuid": chat_guid,
            "message": text,
            "method": opts.method or "apple-script",
        }

        if opts.temp_guid:
            body["tempGuid"] = opts.temp_guid
        if opts.subject:
            body["subject"] = opts.subject
        if opts.effect_id:
            body["effectId"] = opts.effect_id

        data = await self._post("/api/v1/message/text", body)
        message = BlueBubblesMessage(**data)

        logger.info("Sent message: %s", message.guid)

        return SendMessageResult(
            guid=message.guid,
            temp_guid=opts.temp_guid,
            status="sent",
            date_created=message.date_created,
            text=message.text or text,
        )

    async def get_chat(self, chat_guid: str) -> BlueBubblesChat:
        """Gets information about a chat."""
        data = await self._get(f"/api/v1/chat/{quote(chat_guid)}")
        return BlueBubblesChat(**data)

    async def list_chats(
        self, limit: int = 100, offset: int = 0
    ) -> list[BlueBubblesChat]:
        """Lists all chats."""
        data = await self._get(
            f"/api/v1/chat?limit={limit}&offset={offset}&with=lastMessage,participants"
        )
        return [BlueBubblesChat(**chat) for chat in data]

    async def get_messages(
        self, chat_guid: str, limit: int = 50, offset: int = 0
    ) -> list[BlueBubblesMessage]:
        """Gets messages for a chat."""
        data = await self._get(
            f"/api/v1/chat/{quote(chat_guid)}/message?limit={limit}&offset={offset}"
        )
        return [BlueBubblesMessage(**msg) for msg in data]

    async def mark_chat_read(self, chat_guid: str) -> None:
        """Marks a chat as read."""
        await self._post(f"/api/v1/chat/{quote(chat_guid)}/read", {})

    async def react_to_message(
        self, chat_guid: str, message_guid: str, reaction: str
    ) -> None:
        """Sends a reaction to a message."""
        await self._post(
            "/api/v1/message/react",
            {
                "chatGuid": chat_guid,
                "messageGuid": message_guid,
                "reaction": reaction,
            },
        )

    async def edit_message(self, message_guid: str, new_text: str) -> None:
        """Edits a message (requires private API)."""
        await self._post(
            f"/api/v1/message/{quote(message_guid)}/edit",
            {
                "editedMessage": new_text,
                "backwardsCompatibilityMessage": new_text,
            },
        )

    async def unsend_message(self, message_guid: str) -> None:
        """Unsends a message (requires private API)."""
        await self._post(f"/api/v1/message/{quote(message_guid)}/unsend", {})

    async def resolve_target(self, target: str) -> str:
        """Resolves a target to a chat GUID."""
        # If it already looks like a chat GUID, return it
        if target.startswith("iMessage;") or target.startswith("SMS;"):
            return target

        # If it looks like a chat ID, query for it
        if target.startswith("chat_"):
            chats = await self.list_chats(100, 0)
            for chat in chats:
                if (
                    chat.chat_identifier == target
                    or chat.guid == target
                    or target in chat.chat_identifier
                ):
                    return chat.guid

        # Otherwise, construct a DM chat GUID
        return f"iMessage;-;{target}"

    async def create_group_chat(
        self,
        participants: list[str],
        name: str | None = None,
        message: str | None = None,
    ) -> BlueBubblesChat:
        """Creates a new group chat."""
        body: dict[str, Any] = {"participants": participants}
        if name:
            body["name"] = name
        if message:
            body["message"] = message

        data = await self._post("/api/v1/chat", body)
        return BlueBubblesChat(**data)



"""BlueBubbles actions."""

from elizaos_plugin_bluebubbles.actions.send_message import send_message_action
from elizaos_plugin_bluebubbles.actions.send_reaction import send_reaction_action

__all__ = ["send_message_action", "send_reaction_action"]
